"""
Events

This module contains actions for sending various events from a task.
"""
import logging
import os

from typing import TypedDict, Dict, Any, Optional, List

from ppa import events

logger = logging.getLogger(__name__)

SHARED_WORKSPACE = "/workspace/share"

SYSLOG_LEVELS = {"info", "panic", "fatal", "error", "warning", "debug", "trace"}


class Attachment(TypedDict):
    """
- Contains email attachment details

!!! info "All Attachment Keys"

    __name__: The name of the file on the task workspace (including the extension).
    """

    name: str


class WaitHook(TypedDict):
    """
- Contains the output from [_create_wait_hook_](../ppa.events/#create_wait_hook)

- Primarily used to supply parameters to [_wait_](../ppa.events/#wait)

!!! info "All WaitHook Keys"

    __url__: A unique URL assigned to trigger the waithook.

    __secret__: The secret associated with the `url`.
    """

    url: str
    secret: str


class WaitHookResponse(TypedDict):
    """
- Contains the output from [_wait_](../ppa.events/#wait)

!!! info "All WaitHookResponse Keys"

    __username__: The name of the PPA user who responded to the waithook.

    __values__: A dictionary containing the URL parameters from the WaitHook URL responded to.
    """

    username: str
    values: Dict[str, Any]


class ApprovalRequest(TypedDict):
    """
- Generated by the [_create_approval_request_](../ppa.events/#create_approval_request) action

- The `approve_url` & `reject_url` values should be sent to one or more request approvers

!!! info "All ApprovalRequest Keys"

    __approve_url__: A unique URL for approving the request.

    __reject_url__: A unique URL for rejecting the request.

    __secret__: The secret associated with the approval hook.
    """

    approve_url: str
    reject_url: str
    secret: str


class ApprovalResponse(TypedDict):
    """
- Contains the output from [_wait_approval_response_](../ppa.events/#wait_approval_response)

!!! info "All ApprovalResponse Keys"

    __approved__: Indicates whether the request was approved (either `true` or `false`)

    __responder__: The name of the PPA user who responded to the request.
    """

    approved: bool
    responder: str


def send_email(
    subject: str,
    html: str,
    recipients: List[str],
    plain_text: Optional[str] = None,
    reason: Optional[str] = "The task is sending an email",
    attachments: Optional[List[str]] = None,
) -> None:
    """Send an email using the PPA SMTP configuration.

!!! question "Minimum Plugin Version: 4.0.0"

??? info "Input"

    - __subject:__ the email subject

    - __html:__ an HTML document containing the body of the email

    - __plain_text:__ an optional plain text version of the email body for older mail clients

    - __recipients:__ a list of one or more email addresses

    - __attachments:__ optional list of file names to attach to the email

    - __reason:__ text to display when the action runs (defaults to _The task is sending an email_)

!!! tip "Attachments & Recipients"

    **Attachments**

    - The ability to send attachments was introduced in PPA v2.7.0

    - Sending an email will work on older PPA versions, but attempting to send an attachment will fail

    **Recipients**

    - Default SMTP recipients were removed in PPA v2.8.0

    - The `recipients` input to this action is now mandatory

??? success "Output"

    Nothing is outputted by this action.

??? example "Simple HTML Email"
    ```YAML linenums="1"
    - ppa.events.send_email:
        subject: PPA Notification
        html: >
          <html>
            <head>PPA Email Header</head>
            <body>PPA Email Body</body>
          </html>
        plain_text: >
          PPA Email Header

          PPA Email Body
        recipients:
          - recipient_1@email.com
          - recipient_2@email.com
    ```

??? example "Email With Attachment"

    - Auditing a set of users from Active Directory

    - Writing the users to a JSON file

    - Attaching the JSON file to an email

    ```YAML linenums="1"
    - active_directory.groups.get_users:
        distinguishedName: CN=Domain Admins,CN=Users,DC=Example,DC=Domain
      load:
        domain_controller: domain_controller
      save: domain_admins

    - write_json_file:
        name: domain_admins.json
      load:
        contents: domain_admins

    - ppa.events.send_email:
        subject: Domain Admins Report
        html: >
          <html>
            <head>PPA Domain Admins Report</head>
            <body>The JSON report is attached to this email.</body>
          </html>
        attachments:
          - domain_admins.json
        recipients:
          - recipient_1@email.com
          - recipient_2@email.com
    ```

!!! danger "Remember"

    This action will fail if SMTP is not configured in PPA.
    """
    return events.send_email(
        subject=subject,
        html=html,
        recipients=recipients,
        plain_text=plain_text,
        attachments=[
            events.Attachment(**a)
            for a in list(map(lambda a: {"filename": a, "path": os.path.join(SHARED_WORKSPACE, a)}, attachments))
        ]
        if attachments
        else None,
        reason=reason,
    )


def send_syslog_event(
    message: str, level: Optional[str] = "info", reason: Optional[str] = None, details: Optional[Dict[str, Any]] = None,
) -> None:
    """Send a CEF syslog event using the PPA syslog configuration.

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __message:__ the event message

    - __level:__ the event level (defaults to `info`, see more about event levels below)

    - __reason:__ optional text to display in the task interface (defaults to no text displayed)

    - __details:__ optional dictionary of custom event details (defaults to empty)

??? tip "Event Levels"

    Valid event levels:

    - info
    - warning
    - error
    - debug
    - fatal
    - panic
    - trace

??? success "Output"
    Nothing is outputted by this action.

??? example
    {% raw %}
    ```YAML linenums="1"
    - ppa.events.send_syslog_event:
        message: A user's password has been changed by PPA
        details:
          username: "{{ active_directory_user.sAMAccountName }}"
          was_unlocked: false
          force_password_change: true
    ```
    {% endraw %}

??? tip "Generated Syslog Event"

    ```JSON
    {
        "details": {
            "username": "example.user",
            "was_unlocked": false,
            "force_password_change": true
        },
        "level": "info",
        "msg": "A user's password has been changed by PPA",
        "name": "TaskLog",
        "task": "28b54abb-7e9b-4db0-beaa-dad4b0448eeb",
        "time": "2020-06-10T14:51:41Z"
    }
    ```

!!! danger "Remember"

    This action will fail if syslog is not configured in PPA.
    """
    if all([level, level not in SYSLOG_LEVELS]):
        raise ValueError(f"The supplied syslog event level '{level}' is not valid. Level must be one of {SYSLOG_LEVELS}.")
    return events.send_syslog_event(message=message, level=level, reason=reason, details=details)


def create_wait_hook() -> WaitHook:
    """Create a waithook with a URL & secret.

For approval processes consider using the [_create_approval_hook_](#create_approval_hook) instead.

This action allows you to create advanced request processes for your task.

Use in conjunction with the [_wait_](#wait) action.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"
    This action takes no inputs.

??? success "Output"
    A [WaitHook](../inputs_outputs/#waithook) dictionary.

??? example
    ```YAML linenums="1"
    - ppa.events.create_wait_hook:
      save: wait_hook

    # In real life wait_hook.url would be pushed elsewhere at this point...

    - ppa.events.wait:
      load:
        secret: wait_hook.secret
    ```
    """
    url, secret = events.create_wait_hook()
    return WaitHook(url=url, secret=secret)


def wait(text: str, secret: str, element_id: Optional[str] = None) -> WaitHookResponse:
    """Wait for a user to access a waithook URL & authenticate to PPA.

Use in conjunction with the [_create_wait_hook_](#create_wait_hook) action.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __text:__ the text to display while waiting

    - __secret:__ the secret associated with the [WaitHook](../inputs_outputs/#waithook)

    - __element_id:__ an optional ID for the interface element (see [here](../task_interface) for more)

??? success "Output"
    A [WaitHookResponse](../inputs_outputs/#WaitHookResponse) dictionary.

??? example
    - Waiting for a waithook (stored as variable `wait_hook`) to be accessed.

    - Storing the response as a new variable called `response`.

    {% raw %}
    ```YAML linenums="1"
    - ppa.events.create_wait_hook:
      save: wait_hook

    # In real life wait_hook.url would be sent elsewhere at this point...
    - ppa.events.wait:
        text: Waiting for URL {{ wait_hook.url }}...
      load:
        secret: wait_hook.secret
      save: response
    ```
    {% endraw %}

    """
    return WaitHookResponse(**events.wait(text, secret=secret, element_id=element_id))


def create_approval_request() -> ApprovalRequest:
    """Create a waithook with 'approve' & 'reject' URLs & a secret.

This action is similar to [_create_waithook_](#create_wait_hook), but is specifically for approval processes.

Use in conjunction with the [_wait_approval_response_](#wait_approval_response) action.

!!! question "Minimum Plugin Version: 2.1.0"

??? info "Input"
    This action takes no inputs.

??? success "Output"
    An [ApprovalRequest](../inputs_outputs/#approvalrequest) dictionary.

??? example
    ```YAML linenums="1"
    - ppa.events.create_approval_request:
      save: approval_request

    # In real life the approval_request.approve_url & approval_request.reject_url
    # would be pushed somewhere at this point...

    - ppa.events.wait_approval_response:
        text: Waiting for approval response...
        approved_message: The changes will be applied right away!
        rejected_message: The changes will not be applied.
      load:
        approval_request: approval_request
      save: approval_response
    ```
    """
    url, secret = events.create_wait_hook()
    return ApprovalRequest(approve_url=f"{url}?approved", reject_url=f"{url}?rejected", secret=secret)


def wait_approval_response(
    text: str,
    approval_request: ApprovalRequest,
    approved_message: Optional[str] = "",
    rejected_message: Optional[str] = "",
    element_id: Optional[str] = None,
) -> ApprovalResponse:
    """Wait for a user to respond to an approval request.

Use in conjunction with the [_create_approval_request_](#create_approval_request) action.

!!! question "Minimum Plugin Version: 2.1.0"

??? info "Input"

    - __text:__ the text to display while waiting

    - __approval_request:__ an [ApprovalRequest](../inputs_outputs/#approvalrequest)

    - __approved_message:__ an optional message to show the responder if they _approve_ the request

    - __rejected_message:__ an optional message to show the responder if they _reject_ the request

    - __element_id:__ an optional ID for the interface element (see [here](../task_interface) for more)

??? success "Output"

    An [ApprovalResponse](../inputs_outputs/#approvalresponse) dictionary.

??? example

    {% raw %}
    ```YAML linenums="1"
    - ppa.events.create_approval_request:
      save: approval_request

    # In real life the approval_request.approve_url & approval_request.reject_url
    # would be pushed somewhere at this point...

    - ppa.events.wait_approval_response:
        text: Waiting for approval response...
        approved_message: The changes will be applied right away!
        rejected_message: The changes will not be applied.
      load:
        approval_request: approval_request
      save: approval_response
    ```
    {% endraw %}

    """
    response_template = f"""<%#values.approved%>

# Request Approved

__Responder:__ <%username%>

{approved_message}

<%/values.approved%>

<%#values.rejected%>

# Request Rejected

__Responder:__ <%username%>

{rejected_message}

<%/values.rejected%>"""
    response = WaitHookResponse(
        **events.wait(text, secret=approval_request["secret"], element_id=element_id, template=response_template)
    )
    try:
        responder = response["username"].split("\\")[1]
    except IndexError:
        responder = response["username"]
    return ApprovalResponse(responder=responder, approved="approved" in response["values"])
