from typing import TypedDict, Optional, Any

from dateutil import parser

from datacaster.classes import CastDataClass


class UploadedFile(TypedDict):
    """
- Contains data about a file uploaded using [_input_file_](../ppa.ui/#input_file)

!!! info "All UploadedFile Keys"

    __data__: File contents.

    __filename__: Original filename.

    __size__: Size of the file in bytes.

    __type__: Content type (if the browser was able to detect it).
    """

    data: str
    filename: str
    size: int
    type: str


class PayloadKey(TypedDict):
    """
- Describes a key to get from the task payload

!!! info "All PayloadKey Keys"

    __name__: The name of the key in the payload.

    __default_value__: An optional default value to use if the key is missing from the payload.
    """

    name: str
    default_value: Optional[Any] = None


class TimeStamp(TypedDict):
    """
- Contains submitted date picker information

!!! info "All TimeStamp Keys"

    __day__: The selected day.

    __month__: The selected month.

    __year__: The selected year.

    __hour__: The selected hour.

    __minute__: The selected minute.

    __second__: The submitted second.

    __timezone__: The chosen timezone.

    __timestamp__: The supplied date, time, & UTC offset in ISO-8601 format (`yyyy-mm-ddThh:mm:ss+hh:mm`).
    """

    hour: int
    minute: int
    second: int
    day: int
    month: int
    year: int
    timezone: str
    timestamp: str


class Vault(TypedDict):
    """
- Contains information about a Vault configured in PPA

!!! info "All Vault Keys"

    __name__: The name of the Vault.

    __url__: The Vault address.

    __driver__: The type of Vault.

    __expires_at__: The date the Vault access token will expire (if applicable).

!!! example

    ```YAML linenums="1"
    name: AWS Secrets Manager
    url: secretsmanager.eu-west-2.amazonaws.com
    driver: aws-secrets-manager
    expires_at: 09:33:03 03/02/2022
    ```
    """

    name: str
    url: str
    driver: str
    expires_at: str


class VaultData(CastDataClass):
    __annotations__ = Vault.__annotations__
    __class_config__ = {
        "cast_functions": {
            "fields": {
                "url": lambda x: x.replace("https://", "").split(":")[0],
                "expires_at": lambda x: parser.isoparse(x).strftime("%H:%M:%S %d/%m/%Y") if x else "",
            }
        },
        "always_cast": ["url", "expires_at"],
    }
